from collections import OrderedDict
from collections.abc import MutableMapping

try:
    import simplejson as json
    from simplejson import JSONEncoder
except ImportError:
    print("WARNING: Failed to import simplejson, falling back to built-in json")
    import json
    from json import JSONEncoder


def set_missing_key(d, k, v):
    if k not in d:
        d[k] = v


# https://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge
def dictmerge(x, y, path=None, overwrite=False):
    if path is None:
        path = []
    for key in y:
        if key in x:
            if isinstance(x[key], (dict, MutableMapping)) and isinstance(
                y[key], (dict, MutableMapping)
            ):
                dictmerge(x[key], y[key], path + [str(key)], overwrite=overwrite)
            elif x[key] == y[key]:
                pass  # same leaf value
            else:
                if not overwrite:
                    raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
                x[key] = y[key]
        else:
            x[key] = y[key]
    return x


# https://stackoverflow.com/questions/16664874/how-can-i-add-an-element-at-the-top-of-an-ordereddict-in-python
class OrderedDictPlus(OrderedDict):
    def prepend(self, key, value):
        self.update({key: value})
        self.move_to_end(key, last=False)


def _default(self, obj):
    return getattr(obj.__class__, "to_json", _default.default)(obj)


_default.default = JSONEncoder().default
JSONEncoder.default = _default


class JSONMixin:
    # Probably needs a better home
    def to_dict(self):
        if isinstance(self, dict):
            result = self
        else:
            result = self.__dict__.copy()
        for k, v in result.items():
            if hasattr(v, "to_dict"):
                result[k] = v.to_dict()
        return result

    # This is used for _defaults in JSON encoding
    def to_json(self):
        return self.__dict__

    def to_jsons(self):
        return json.dumps(self.__dict__)



from functools import reduce
from itertools import chain, combinations
from operator import or_

from orderedset import OrderedSet


def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def chunks(l, chunksize):
    for i in range(0, len(l), chunksize):
        yield l[i : i + chunksize]


def nchunks(a, n):
    """Divide iterable a into n chunks"""
    k, m = divmod(len(a), n)
    return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))


def read_chunks(reader, chunksize, limit=None):
    """Given an iterator/generator, read/yield chunks of chunksize up to
    (optional) limit"""
    chunk = []
    for i, line in enumerate(reader):
        if limit and i >= limit:
            break
        if i % chunksize == 0 and i > 0:
            yield chunk
            chunk = []
        chunk.append(line)
    yield chunk


def iter_or(iterable):
    return reduce(or_, iterable)


def orderedsetify(obj):
    """Take ordered iterable and turn it into OrderedSet"""
    if isinstance(obj, OrderedSet):
        return obj
    if isinstance(obj, (list, tuple)):
        return OrderedSet(obj)
    assert False, "Not sure how to setify %s" % obj



import time

import requests

from tlbx.logging_utils import dbg


def poll_call(
    func, result_param, result_value, sleep_time, max_iter, *_args, **_kwargs
):
    i = 0
    while True:
        result = func(*_args, **_kwargs)
        if result[result_param] == result_value:
            return result
        i += 1
        assert i < max_iter, (
            "Exhausted poll_call, no result in %d tries. Last result: %s"
            % (max_iter, result)
        )
        dbg(
            "Polling %s, iteration %d/%d, %s=%s"
            % (func.__name__, i, max_iter, result_param, result[result_param])
        )
        if sleep_time:
            time.sleep(sleep_time)


def paged_call(func, size_param, offset_param, page_size, *_args, **_kwargs):
    offset = 0
    results = []

    while True:
        _kwargs.update({size_param: page_size, offset_param: offset})
        result = func(*_args, **_kwargs)
        results.extend(result)
        result_len = len(result)
        offset += result_len
        if result_len < page_size:
            break

    print("Got %d results" % offset)
    return results


def paged_get(url, size_param, offset_param, page_size, *_args, **_kwargs):
    offset = 0
    results = []

    while True:
        _kwargs["params"] = _kwargs.get("params", {})
        _kwargs["params"].update({size_param: page_size, offset_param: offset})
        resp = requests.get(url, *_args, **_kwargs)
        resp.raise_for_status()
        result = resp.json()
        results.extend(result)
        result_len = len(result)
        offset += result_len
        if result_len < page_size:
            break

    print("Got %d results" % offset)
    return results



import logging
from pprint import pformat as pf, pprint as pp
from reprlib import Repr
import shutil
import sys
from textwrap import indent, wrap

import sqlparse as sp

from tlbx.object_utils import get_class_vars, get_caller


logging.basicConfig(format="%(message)s")


_repr = Repr()
_repr.maxstring = 60
_repr.maxother = 20
_repr.maxlist = 5
_repr.maxtuple = 5
_repr.maxset = 5
repr = _repr.repr


class FontSpecialChars:
    ENDC = "\033[0m"


class FontColors:
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    NONE = ""  # Will use default terminal coloring


RESERVED_COLORS = [
    "RED",  # errors
    "YELLOW",  # warnings
    "BLACK",  # to avoid conflicts with terminal defaults
    "WHITE",  # to avoid conflicts with terminal defaults
    "NONE",  # to avoid conflicts with terminal defaults
]

COLOR_OPTIONS = sorted(
    [x for x in get_class_vars(FontColors) if x not in RESERVED_COLORS]
)


class FontEffects:
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    INVERTED = "\033[7m"


def get_terminal_width():
    return shutil.get_terminal_size().columns


def get_logging_level(logger=None):
    if (not logger) or (not isinstance(logger, logging.Logger)):
        logger = logging.getLogger()
    return logger.getEffectiveLevel()


def sqlformat(sql):
    return sp.format(sql, reindent=True, keyword_case="upper")


def format_msg(
    msg, label="parent", indent=None, color=None, autocolor=False, format_func=pf
):
    term_width = get_terminal_width()
    effective_width = term_width

    if label == "parent":
        label = get_caller()

    if indent is not None:
        if indent == "label":
            assert label, "indent=label but no label given"
            indent_str = " " * (len(label) + 2)
        else:
            indent = int(indent)
            indent_str = " " * indent
        effective_width = effective_width - len(indent_str)

    if isinstance(format_func, str):
        format_func = globals()[format_func]

    if format_func and format_func != pf:
        msg = format_func(msg)
        if label and not indent:
            msg = "\n" + msg
    elif not isinstance(msg, str):
        msg = pf(msg, width=effective_width)
        if label and not indent:
            msg = "\n" + msg

    if indent is not None:
        lines = msg.splitlines(True)
        msg = "\n".join(["\n".join(wrap(line, effective_width)) for line in lines])
        if indent == "label":
            msg = indent_str.join(msg.splitlines(True))
        else:
            msg = indent_str + indent_str.join(msg.splitlines(True))
            if label:
                msg = "\n" + msg

    if label:
        msg = label.strip() + ": " + msg

    if (not color) and autocolor:
        assert label, "No label provided, can not use autocolor"
        color_index = ord(label[0]) % len(COLOR_OPTIONS)
        color = COLOR_OPTIONS[color_index]

    if color:
        msg = getattr(FontColors, color.upper()) + msg + FontSpecialChars.ENDC

    return msg


def info(msg, label="parent", logger=None, autocolor=True, **kwargs):
    logger = logger or logging
    if get_logging_level(logger) > logging.INFO:
        return
    if label == "parent":
        label = get_caller()
    msg = format_msg(msg, label=label, autocolor=autocolor, **kwargs)
    logger.info(msg)


def dbg(msg, label="parent", logger=None, autocolor=True, **kwargs):
    logger = logger or logging
    if get_logging_level(logger) > logging.DEBUG:
        return
    if label == "parent":
        label = get_caller()
    msg = format_msg(msg, label=label, autocolor=autocolor, **kwargs)
    logger.debug(msg)


def dbgsql(msg, label="parent", logger=None, autocolor=True, **kwargs):
    logger = logger or logging
    if get_logging_level(logger) > logging.DEBUG:
        return
    if label == "parent":
        label = get_caller()
    msg = format_msg(
        msg, label=label, autocolor=autocolor, format_func=sqlformat, **kwargs
    )
    logger.debug(msg)


def warn(msg, label="parent", logger=None, **kwargs):
    logger = logger or logging
    if get_logging_level(logger) > logging.WARNING:
        return
    if label == "parent":
        label = get_caller()
    msg = format_msg(msg, label=label, color="yellow", **kwargs)
    logger.warning(msg)


def error(msg, label="parent", logger=None, **kwargs):
    logger = logger or logging
    if get_logging_level(logger) > logging.ERROR:
        return
    if label == "parent":
        label = get_caller()
    msg = format_msg(msg, label=label, color="red", **kwargs)
    logger.error(msg)


class PrintMixin:
    repr_attrs = []

    def __repr__(self):
        if self.repr_attrs:
            return "<%s %s>" % (
                type(self).__name__,
                " ".join(
                    [
                        "%s=%s" % (field, repr(getattr(self, field)))
                        for field in self.repr_attrs
                    ]
                ),
            )
        return "<%s %s>" % (type(self).__name__, id(self))

    def __str__(self):
        return str(pf(vars(self)))



import random
import string


def is_str(o):
    try:
        basestring
    except NameError:
        basestring = (str, bytes)
    return isinstance(o, basestring)


def is_int(v):
    if v is True or v is False:
        return False

    try:
        int(str(v))
        return True
    except ValueError:
        return False


def is_num(v):
    if v is True or v is False:
        return False

    try:
        float(v)
        return True
    except ValueError:
        return False


def get_string_format_args(s):
    return [tup[1] for tup in string.Formatter().parse(s) if tup[1] is not None]


def string_has_format_args(s):
    if get_string_format_args(s):
        return True
    return False


def random_string(length=10):
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))



import inspect

import climax


Parent = climax.parent


def st():
    __import__("pdb").Pdb().set_trace(inspect.currentframe().f_back)


class Arg:
    """Decorator around climax.argument that stores arg values"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.climax_arg = climax.argument(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.climax_arg(*args, **kwargs)

    @property
    def name(self):
        return self.args[0]


class Script:
    """Decorator for main function in a script that provides CLI parsing"""

    parents = []  # Override as list of climax.parent objects to inherit args

    def __init__(self, *script_args):
        self.func = None
        self.script_args = script_args or []

    def get_injected_kwargs(self):
        return {}

    def clean_up(self, **kwargs):
        pass

    def __call__(self, func, *args, **kwargs):
        self.func = func

        def inner(*args, **kwargs):
            kwargs.update(self.get_injected_kwargs())
            try:
                result = func(*args, **kwargs)
            finally:
                self.clean_up(**kwargs)
            return result

        for script_arg in self.script_args:
            inner = script_arg(inner)
        inner = climax.command(parents=self.parents)(inner)
        return inner


def prompt_user(msg, answers):
    answer = None
    answers = [str(x).lower() for x in answers]
    display_answers = "[%s] " % "/".join(answers)
    while (answer is None) or (answer.lower() not in answers):
        answer = input("%s %s" % (msg, display_answers))
    return answer



from collections.abc import MutableMapping, Callable
from functools import wraps
from importlib import import_module
import inspect
import sys


def get_class_vars(cls):
    return {
        i for i in dir(cls) if (not isinstance(i, Callable)) and (not i.startswith("_"))
    }


def get_class_var_values(cls):
    return {
        getattr(cls, i)
        for i in dir(cls)
        if (not isinstance(i, Callable)) and (not i.startswith("_"))
    }


class ClassVarContainsMeta(type):
    def __contains__(cls, key):
        return key in get_class_vars(cls)


class ClassValueContainsMeta(type):
    def __contains__(cls, key):
        return key in get_class_var_values(cls)


def import_object(name):
    if "." not in name:
        frame = sys._getframe(1)
        module_name = frame.f_globals["__name__"]
        object_name = name
    else:
        module_name = ".".join(name.split(".")[:-1])
        object_name = name.split(".")[-1]
    return getattr(import_module(module_name), object_name)


# https://stackoverflow.com/questions/1389180/automatically-initialize-instance-variables
def initializer(func):
    names, varargs, keywords, defaults = inspect.getargspec(func)

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        for name, arg in list(zip(names[1:], args)) + list(kwargs.items()):
            setattr(self, name, arg)
        if defaults:
            for i in range(len(defaults)):
                index = -(i + 1)
                if not hasattr(self, names[index]):
                    setattr(self, names[index], defaults[index])
        func(self, *args, **kwargs)

    return wrapper


class MappingMixin(MutableMapping):
    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


# TODO: needs a better home?
def get_caller(depth=0):
    # Add 1 to depth so its the caller of the caller to this function
    return sys._getframe(depth + 1).f_back.f_code.co_name



import os
import subprocess

from pandas.io.common import get_filepath_or_buffer

try:
    from pandas.io.common import _get_handle as get_handle
except ImportError:
    from pandas.io.common import get_handle


def rmfile(filename, ignore_missing=False):
    try:
        os.remove(filename)
    except FileNotFoundError:
        if ignore_missing:
            return
        raise


def shell(cmd, shell=True, capture_output=True, **kwargs):
    return subprocess.run(cmd, shell=shell, capture_output=capture_output, **kwargs)


def open_filepath_or_buffer(f, open_flags="r", compression=None):
    """Use pandas IO functions to return a handle from a filepath
    or buffer.

    Parameters
    ----------
    f : str or buffer
        filepath or buffer to open
    open_flags : str, optional
        mode to open file
    compression : str, optional
        compression arg passed to pandas functions

    Returns
    -------
    f : file-like
        A file-like object
    handles : list of file-like
        A list of file-like objects opened. Seems mostly relevant for zipped archives.
    close : bool
        A flag indicating whether the caller should close the file object when done

    """
    f, _, compression, should_close = get_filepath_or_buffer(f, compression=compression)

    close = False or should_close
    if isinstance(f, str):
        close = True

    f, handles = get_handle(f, open_flags, compression=compression)

    return f, handles, close



import attr
from email import parser, policy
from email.message import EmailMessage
import logging
import mimetypes
import ntpath
import smtplib

from imapclient import IMAPClient

from tlbx.object_utils import MappingMixin

SMTP_SSL_PORT = 465

imapclient_logger = logging.getLogger("imapclient")
imapclient_logger.setLevel(logging.WARNING)


@attr.s(kw_only=True)
class EmailPayload(MappingMixin):
    payload = attr.ib()
    content_type = attr.ib()
    charset = attr.ib()


def update_email(
    msg, frm=None, to=None, subject=None, body=None, html=None, attachments=None
):
    if frm:
        msg.replace_header("from", frm)

    if to:
        msg.replace_header("to", to)

    if subject:
        msg.replace_header("subject", subject)

    if body:
        if msg.is_multipart():
            payload = msg.get_payload()
            new_body = EmailMessage()
            new_body.set_content(body)
            payload[0] = new_body
            msg.set_payload(payload)
        else:
            msg.set_content(body)

    if html:
        msg.add_alternative(html, subtype="html")

    if attachments:
        if msg.is_multipart():
            payload = msg.get_payload()
            # Try to only keep the body. There might be a better way
            msg.set_payload(payload[:1])
        add_email_attachments(attachment)

    return msg


# -------- Read Emails


def extract_email_payload(msg, decode=True):
    def decode_if_necessary(payload, charset):
        # decode=True still returns encoded bytes, but just decodes
        # the message contents from transfer encoding, such as if the
        # payload is base64 encoded.
        charset = charset or "utf8"  # Fall back to utf8
        if (not decode) or (not isinstance(payload, bytes)):
            return payload
        return payload.decode(charset)

    if msg.is_multipart():
        parts = []
        for part in msg.walk():
            if "multipart" in part.get_content_type().lower():
                continue
            charset = part.get_content_charset()
            payload = part.get_payload(decode=decode)
            payload = decode_if_necessary(payload, charset)
            parts.append(
                EmailPayload(
                    content_type=part.get_content_type(),
                    charset=charset,
                    payload=payload,
                )
            )
        return parts
    else:
        # When is_multipart is False, payload is a string
        charset = msg.get_content_charset()
        payload = msg.get_payload(decode=decode)
        payload = decode_if_necessary(payload, charset)
        return [
            EmailPayload(
                content_type=msg.get_content_type(), charset=charset, payload=payload
            )
        ]


def read_email(
    criteria,
    folder="INBOX",
    sort=None,
    limit=None,
    client=None,
    host=None,
    username=None,
    password=None,
    **kwargs
):
    results = []
    logout = False

    if not client:
        assert (
            host and username and password
        ), "Must pass host/username/password for IMAP connection"
        logout = True
        client = IMAPClient(host, **kwargs)
        client.login(username, password)

    try:
        select_info = client.select_folder(folder)
        if sort:
            messages = client.sort(sort, criteria=criteria)
        else:
            messages = client.search(criteria)

        if limit:
            messages = messages[:limit]

        for msg_id, msg_data in client.fetch(messages, ["RFC822"]).items():
            raw = msg_data[b"RFC822"].decode("utf8")
            msg = parser.Parser(policy=policy.default).parsestr(raw)
            results.append(msg)

    finally:
        if logout:
            client.logout()

    return results


# -------- Write Emails


def add_email_attachments(msg, attachments):
    for attachment in attachments or []:
        assert ntpath.isfile(attachment), "Email attachments must be valid file paths"
        filename = ntpath.basename(attachment)

        ctype, encoding = mimetypes.guess_type(attachment)
        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)

        with open(attachment, "rb") as fp:
            msg.add_attachment(
                fp.read(), maintype=maintype, subtype=subtype, filename=filename
            )


def create_email(frm, to, subject, body=None, html=None, attachments=None):
    msg = EmailMessage()
    msg["From"] = frm
    msg["To"] = to if isinstance(to, str) else ", ".join(to)
    msg["Subject"] = subject
    msg.preamble = "You will not see this in a MIME-aware mail reader.\n"
    msg.set_content(body or "")

    if html is not None:
        msg.add_alternative(html, subtype="html")

    add_email_attachments(msg, attachments)
    return msg


def send_email(
    msg,
    client=None,
    host=None,
    port=None,
    username=None,
    password=None,
    smtp_class=smtplib.SMTP,
    debug=False,
):
    if client:
        client.send_message(msg)
        return

    assert (
        host and port and username and password
    ), "Must pass host/port/username/password for SMTP connection"

    if port == SMTP_SSL_PORT:
        smtp_class = smtplib.SMTP_SSL

    with smtp_class(host, port=port) as client:
        client.ehlo()
        if smtp_class == smtplib.SMTP:
            client.starttls()
            client.ehlo()
        client.login(username, password)
        client.set_debuglevel(debug)
        client.send_message(msg)



#!/usr/bin/env python

from contextlib import contextmanager
import os

from setuptools import find_packages, setup


@contextmanager
def load_file(fname):
    f = open(os.path.join(os.path.dirname(__file__), fname))
    try:
        yield f
    finally:
        f.close()


with load_file("README.md") as f:
    README = f.read()

with load_file("requirements.txt") as f:
    requires = f.read().split("\n")


exec(open("tlbx/version.py").read())

setup(
    name="tlbx",
    description="Just some common utilities.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/totalhack/tlbx",
    author="totalhack",
    author_email="none@none.com",
    maintainer="totalhack",
    version=__version__,
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    install_requires=requires,
)



from tlbx.callable_utils import *
from tlbx.cli_utils import *
from tlbx.date_utils import *
from tlbx.dict_utils import *
from tlbx.email_utils import *
from tlbx.file_utils import *
from tlbx.iterable_utils import *
from tlbx.logging_utils import *
from tlbx.object_utils import *
from tlbx.string_utils import *
